#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include "base64.h"

#define END 126
#define NAN 127
// Base64 table
static const char base64_table[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
static const char base64_decode_table[128] = {
	NAN, NAN, NAN, NAN, NAN, NAN, NAN, NAN,
	NAN, NAN, NAN, NAN, NAN, NAN, NAN, NAN,
	NAN, NAN, NAN, NAN, NAN, NAN, NAN, NAN,
	NAN, NAN, NAN, NAN, NAN, NAN, NAN, NAN,
	NAN, NAN, NAN, NAN, NAN, NAN, NAN, NAN,
	NAN, NAN, NAN,  62, NAN, NAN, NAN,  63,
	 52,  53,  54,  55,  56,  57,  58,  59,
	 60,  61, NAN, NAN, NAN, END, NAN, NAN,
	NAN,   0,   1,   2,   3,   4,   5,   6,
	  7,   8,   9,  10,  11,  12,  13,  14,
	 15,  16,  17,  18,  19,  20,  21,  22,
	 23,  24,  25, NAN, NAN, NAN, NAN, NAN,
	NAN,  26,  27,  28,  29,  30,  31,  32,
	 33,  34,  35,  36,  37,  38,  39,  40,
	 41,  42,  43,  44,  45,  46,  47,  48,
	 49,  50,  51, NAN, NAN, NAN, NAN, NAN
};

// Base64 encode return 0 if failed.
int encode_base64(
    char *const encoded_data, const unsigned char* data, 
    size_t input_length)
{
    if (
        encoded_data == NULL ||
        data == NULL
    ) {
        return 0;
    }

    char* running_encoded_data = encoded_data;
    uint32_t encode_a, encode_b, encode_c, combined;
    size_t i;

    for (i = 0; i + 2 < input_length;) {
        encode_a = data[i++];
        encode_b = data[i++];
        encode_c = data[i++];

        combined = (encode_a << 0x10) + (encode_b << 0x08) + encode_c;

        *running_encoded_data++ = base64_table[(combined >> (3 * 6)) & 0x3F];
        *running_encoded_data++ = base64_table[(combined >> (2 * 6)) & 0x3F];
        *running_encoded_data++ = base64_table[(combined >> (1 * 6)) & 0x3F];
        *running_encoded_data++ = base64_table[(combined >> (0 * 6)) & 0x3F];
    }

    if (input_length - i > 0) {
        encode_a = data[i];
        if (input_length - i == 1) {
            combined = encode_a << 0x10;
            *running_encoded_data++ = base64_table[(combined >> (3 * 6)) & 0x3F];
            *running_encoded_data++ = base64_table[(combined >> (2 * 6)) & 0x3F];
            *running_encoded_data++ = '=';
        } else {
            encode_b = data[i + 1];
            combined = (encode_a << 0x10) + (encode_b << 0x08);
            *running_encoded_data++ = base64_table[(combined >> (3 * 6)) & 0x3F];
            *running_encoded_data++ = base64_table[(combined >> (2 * 6)) & 0x3F];
            *running_encoded_data++ = base64_table[(combined >> (1 * 6)) & 0x3F];
        }
        *running_encoded_data++ = '=';
    }

    *running_encoded_data = '\0';
    return 1;
}

// Base64 decode return 0 if failed.
int decode_base64(
    unsigned char* const decoded_data, const char* data, 
    size_t input_length, size_t output_max_length, size_t *output_length)
{
    if (
        (input_length & 0x3) != 0 || 
        decoded_data == NULL    || 
        data == NULL
    ) {
        return 0;
    }

    size_t output_length_check = (input_length >> 2) * 3;
    if (output_length_check <= 0) {   
        return 0;
    }
    
    if (data[input_length - 1] == '=') output_length_check--;
    if (data[input_length - 2] == '=') output_length_check--;
    if (output_length_check > output_max_length) {
        return 0;
    }

    unsigned char* running_decoded_data = decoded_data;
    const unsigned char* udata = (const unsigned char*)data;

    uint32_t decode_a, decode_b, decode_c, decode_d, combined;

    for (size_t i = 0; i < input_length;) {
        decode_a = base64_decode_table[udata[i++]];
        decode_b = base64_decode_table[udata[i++]];
        decode_c = base64_decode_table[udata[i++]];
        decode_d = base64_decode_table[udata[i++]];

        if (
            decode_a == NAN ||
            decode_a == END ||
            decode_b == NAN ||
            decode_b == END ||
            decode_c == NAN ||
            decode_d == NAN
        ) {
            memset(decoded_data, 0, (size_t)(running_decoded_data - decoded_data));
            if (output_length != NULL) {
                *output_length = 0;
            }
            return 0;
        }

        if (decode_c == END || decode_d == END) {
            break;
        }

        combined = (decode_a << (3 * 6)) +
                   (decode_b << (2 * 6)) +
                   (decode_c << (1 * 6)) +
                   (decode_d << (0 * 6));

        *running_decoded_data++ = (combined >> (2 * 8)) & 0xFF;
        *running_decoded_data++ = (combined >> (1 * 8)) & 0xFF;
        *running_decoded_data++ = (combined >> (0 * 8)) & 0xFF;
    }

    combined = (decode_a << (3 * 6)) + 
               (decode_b << (2 * 6));

    if (decode_d == END) {
        combined = (decode_c != END) ? (combined + (decode_c << (1 * 6))) : combined;
        *running_decoded_data++ = (combined >> (2 * 8)) & 0xFF;
        if (decode_c != END) {
            *running_decoded_data++ = (combined >> (1 * 8)) & 0xFF;
        } 
    }

    if (output_length != NULL) {
        *output_length = (size_t)(running_decoded_data - decoded_data);
    }

    return 1;
}
