#include <getopt.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include "base64.h"
#include "utils.h"
#include "pqsign.h"
#include "sm3sum/sm3sum.h"

int sign_flag = 0;
int genkey_flag = 0;
int verify_flag = 0;
int free_flag = 0;

/// TODO: Add sk passwd protect.

static void usage(void) {
    puts(
        "pqsign version " PQSIGN_VERSION " use algorithm " ALG_NAME "\n"
        "\nUsage:\n"
        "pqsign -G [-p pubkey_file] [-k seckey_file]\n"
        "pqsign -S <file> [-p <pubkey_file>] [-x <sig_file>]\n"
        "pqsign -V <file> [-k <seckey_file>] [-x <sig_file>]\n"
        "\nSign or verify a file using Aigis-sig.\n"
        "\n"
        "-G, --genkey                   generate a new key pair\n"
        "-S, --sign    <file>           sign the specified file\n"
        "-V, --verify  <file>           verify the signature of the specified file\n"
        "-p, --pubkey  <pubkey_file>    public key file (default: ./pqsign.pub)\n"
        "-k, --seckey  <seckey_file>    secret key file (default: ./pqsign.key)\n"
        "-x, --sigfile <sig_file>       signature file (default: <file>.pqsig)\n"
        "-h, --help                     display this help message and exit\n"
        "    --version                  output version information\n"
    );
}

static void write_pk_to_file(const char *pk_file, const PubKey *pubkey) 
{
    char write_content[BASE64_ENCODE_MAX_LEN(ALG_NAME_MAX_SIZE + SIG_PUBLICKEYBYTES)];
    memset(write_content, 0, BASE64_ENCODE_MAX_LEN(ALG_NAME_MAX_SIZE + SIG_PUBLICKEYBYTES));

    if (
        encode_base64(
            write_content, (const unsigned char*)pubkey, 
            ALG_NAME_MAX_SIZE + SIG_PUBLICKEYBYTES) == 0
    ) {
        exit_with_error_msg("Encode public key error");
    }

    if (
        write_file(pk_file, write_content, strlen(write_content)) < 0
    ) {
        exit_with_error_msg("Write public key file error");
    }

}

static void load_pk_from_file(const char *pk_file, const PubKey *pubkey) 
{
    char read_content[BASE64_ENCODE_MAX_LEN(ALG_NAME_MAX_SIZE + SIG_PUBLICKEYBYTES)];
    memset(read_content, 0, BASE64_ENCODE_MAX_LEN(ALG_NAME_MAX_SIZE + SIG_PUBLICKEYBYTES));
    
    int read_size = read_file(
                        pk_file, read_content, 
                        BASE64_ENCODE_MAX_LEN(ALG_NAME_MAX_SIZE + SIG_PUBLICKEYBYTES));
    if (read_size < 0) {
        if (
            read_size == PQSIGN_FILE_OPEN_ERROR || 
            read_size == PQSIGN_FILE_SIZE_ERROR) 
        {
            fprintf(stderr, "Please confirm if the public key file is valid.\n");
        }
        exit_with_error_msg("Load public key error");
    }

    if (
        decode_base64(
            (unsigned char*)pubkey, read_content, 
            read_size, 
            ALG_NAME_MAX_SIZE + SIG_PUBLICKEYBYTES, 
            NULL) == 0
    ) {
        exit_with_error_msg("Decode public key error");
    }

}

static void write_sk_to_file(const char *sk_file, const SecKey *seckey) 
{
    char write_content[BASE64_ENCODE_MAX_LEN(ALG_NAME_MAX_SIZE + SIG_SECRETKEYBYTES)];
    memset(write_content, 0, BASE64_ENCODE_MAX_LEN(ALG_NAME_MAX_SIZE + SIG_SECRETKEYBYTES));

    if (
        encode_base64(
            write_content, (const unsigned char*)seckey, 
            ALG_NAME_MAX_SIZE + SIG_SECRETKEYBYTES) == 0
    ) {
        exit_with_error_msg("Encode secret key error");
    }

    if (
        write_file(sk_file, write_content, strlen(write_content)) < 0
    ) {
        exit_with_error_msg("Write secret key file error");
    }

}

static void load_sk_from_file(const char *sk_file, const SecKey *seckey) 
{
    char read_content[BASE64_ENCODE_MAX_LEN(ALG_NAME_MAX_SIZE + SIG_SECRETKEYBYTES)];
    memset(read_content, 0, BASE64_ENCODE_MAX_LEN(ALG_NAME_MAX_SIZE + SIG_SECRETKEYBYTES));
    
    int read_size = read_file(
                        sk_file, read_content, 
                        BASE64_ENCODE_MAX_LEN(ALG_NAME_MAX_SIZE + SIG_SECRETKEYBYTES));
    if (read_size < 0) {
        if (
            read_size == PQSIGN_FILE_OPEN_ERROR || 
            read_size == PQSIGN_FILE_SIZE_ERROR) 
        {
            fprintf(stderr, "Please confirm if the secret key file is valid.\n");
        }
        exit_with_error_msg("Load secret key error");
    }

    if (
        decode_base64(
            (unsigned char*)seckey, read_content, 
            read_size, 
            ALG_NAME_MAX_SIZE + SIG_SECRETKEYBYTES, 
            NULL) == 0
    ) {
        exit_with_error_msg("Decode secret key error");
    }

}

static void write_sig_to_file(const char *sig_file, const Sig *sig) 
{
    char write_content[BASE64_ENCODE_MAX_LEN(ALG_NAME_MAX_SIZE + SIG_BYTES + sizeof(size_t)*8)];
    memset(write_content, 0, BASE64_ENCODE_MAX_LEN(ALG_NAME_MAX_SIZE + SIG_BYTES + sizeof(size_t)*8));
    
    if (
        encode_base64(
            write_content, (const unsigned char*)sig, 
            ALG_NAME_MAX_SIZE + SIG_BYTES + sizeof(size_t)*8) == 0
    ) {
        exit_with_error_msg("Encode sig error");
    }

    if (
        write_file(sig_file, write_content, strlen(write_content)) < 0
    ) {
        exit_with_error_msg("Write sig file error");
    }

}

static void load_sig_from_file(const char *sig_file, const Sig *sig) 
{
    char read_content[BASE64_ENCODE_MAX_LEN(ALG_NAME_MAX_SIZE + SIG_BYTES + sizeof(size_t)*8)];
    memset(read_content, 0, BASE64_ENCODE_MAX_LEN(ALG_NAME_MAX_SIZE + SIG_BYTES + sizeof(size_t)*8));
    
    int read_size = read_file(
                        sig_file, read_content, 
                        BASE64_ENCODE_MAX_LEN(ALG_NAME_MAX_SIZE + SIG_BYTES + sizeof(size_t)*8));
    if (read_size < 0) {
        if (
            read_size == PQSIGN_FILE_OPEN_ERROR || 
            read_size == PQSIGN_FILE_SIZE_ERROR) 
        {
            fprintf(stderr, "[ERROR] Please confirm if the sig file is valid.\n");
        }
        exit_with_error_msg("Load sig error");
    }

    if (
        decode_base64(
            (unsigned char*)sig, read_content, 
            read_size, 
            ALG_NAME_MAX_SIZE + SIG_BYTES + sizeof(size_t)*8, 
            NULL) == 0
    ) {
        exit_with_error_msg("Decode sig error");
    }

}


static int generate(const char *pk_file, const char *sk_file) 
{
    PubKey pubkey;
    SecKey seckey;

    memset(&pubkey, 0, sizeof(PubKey));
    memset(&seckey, 0, sizeof(SecKey));

    crypto_sign_keypair(pubkey.pk, seckey.sk);
    memcpy(pubkey.alg_name, ALG_NAME, strlen(ALG_NAME));
    memcpy(seckey.alg_name, ALG_NAME, strlen(ALG_NAME));

    write_pk_to_file(pk_file, &pubkey);
    write_sk_to_file(sk_file, &seckey);

    return 0;
}

static int sign_hash_str(
    SecKey *seckey, 
    const unsigned char *hash_str, size_t hash_str_len, 
    const char *sig_file)
{
    Sig sig;
    memset(&sig, 0, sizeof(Sig));
    memcpy(sig.alg_name, ALG_NAME, strlen(ALG_NAME));

    if (
        crypto_sign_signature(
            sig.sig, &sig.sig_len, 
            hash_str, hash_str_len, 
            seckey->sk) < 0
    ) {
        return 0;
    }

    write_sig_to_file(sig_file, &sig);

    return 1;
}

static int sign_file(
    SecKey *seckey, const char *filename, const char *sig_file)
{
    sm3_file_ctx ctx;
    strcpy(ctx.filename, filename);
    calc_file_hash(&ctx);

    return sign_hash_str(
                seckey, 
                (const unsigned char*)ctx.calculated_sm3, 
                SM3_DIGEST_LENGTH, 
                sig_file);
}

static int verify_hash_str(
    PubKey *pubkey, 
    const unsigned char *hash_str, size_t hash_str_len, 
    const char *sig_file)
{
    Sig sig;
    memset(&sig, 0, sizeof(Sig));

    load_sig_from_file(sig_file, &sig);

    if (
        crypto_sign_verify(
            sig.sig, sig.sig_len, 
            hash_str, hash_str_len, 
            pubkey->pk) <= 0
    ) {
        return 0;
    }

    return 1;
}


static int verify_file(
    PubKey *pubkey, const char *filename, const char *sig_file)
{
    sm3_file_ctx ctx;
    strcpy(ctx.filename, filename);
    calc_file_hash(&ctx);

    return verify_hash_str(
                pubkey, 
                (const unsigned char*)ctx.calculated_sm3, 
                SM3_DIGEST_LENGTH, 
                sig_file);
}


static char *append_sig_extension(const char *filename)
{
    char  *sig_file;
    size_t filename_len = strlen(filename);

    sig_file = (char*)malloc(filename_len + strlen(DEFAULT_SIG_EXT) + 1);
    if (!sig_file) {
        exit_with_perror_msg("[ERROR] malloc sig_file");
    }
    memcpy(sig_file, filename, filename_len);
    memcpy(sig_file + filename_len, DEFAULT_SIG_EXT, strlen(DEFAULT_SIG_EXT));
    sig_file[filename_len + strlen(DEFAULT_SIG_EXT)] = '\0';

    return sig_file;
}

static void check_conflicts() {
    if ((sign_flag + genkey_flag + verify_flag) > 1) {
        fprintf(stderr, "[ERROR] Only one of -s, -g, or -v can be specified at a time.\n");
        usage();
        exit(EXIT_FAILURE);
    }
}


int main(int argc, char *argv[]) {
    int opt;
    int option_index = 0;
    char *sign_filename = NULL;
    char *verify_filename = NULL;
    char *pubkey_filename = NULL;
    char *seckey_filename = NULL;
    char *signature_filename = NULL;

    struct option long_options[] = {
        {"sign",    required_argument, 0, 'S'},
        {"genkey",  no_argument,       0, 'G'},
        {"verify",  required_argument, 0, 'V'},
        {"pubkey",  required_argument, 0, 'p'},
        {"seckey",  required_argument, 0, 'k'},
        // {"output",  required_argument, 0, 'o'},
        {"sigfile", required_argument, 0, 'x'},
        {"help",    no_argument,       0, 'h'},
        {"version", no_argument,       0,  0 },
        {0, 0, 0, 0}
    };

    if (argc < 2) {
        fprintf(stderr, "%s: need an option.\n", argv[0]);
        usage();
        exit(EXIT_FAILURE);
    }

    while ((opt = getopt_long(argc, argv, "S:GV:p:k:x:h", long_options, &option_index)) != -1) {
        switch (opt) {
            case 'S':
                sign_filename = optarg;
                sign_flag += 1;
                check_conflicts();
                break;
            case 'p':
                pubkey_filename = optarg;
                break;
            case 'k':
                seckey_filename = optarg;
                break;
            case 'G':
                genkey_flag += 1;
                check_conflicts();
                break;
            // case 'o':
            //     signature_filename = optarg;
            //     break;
            case 'V':
                verify_filename = optarg;
                verify_flag += 1;
                check_conflicts();
                break;
            case 'x':
                signature_filename = optarg;
                break;
            case 'h':
                usage();
                exit(EXIT_SUCCESS);
            case 0:
                if (strncmp(long_options[option_index].name, "version", 8) == 0) {
                    printf("pqsign version %s use algorithm %s\n", PQSIGN_VERSION, ALG_NAME);
                    exit(EXIT_SUCCESS);
                } else{
                    fprintf(stderr, "%s: invaild opttion -- '%s'\n", argv[0], long_options[option_index].name);
                    usage();
                    exit(EXIT_FAILURE);
                }
            case '?':
                usage();
                exit(EXIT_FAILURE);
            default:
                usage();
                exit(EXIT_FAILURE);
        }
    }

    if (genkey_flag) {
        if (!pubkey_filename) {
            pubkey_filename = DEFAULT_PUBKEY_FILE;
        }
        if (!seckey_filename) {
            seckey_filename = DEFAULT_SECKEY_FILE;
        }

        if (file_exists(pubkey_filename)) {
            fprintf(stderr, "[ERROR] Public key %s exists.\n", pubkey_filename);
            exit(EXIT_FAILURE);
        } 
        if (file_exists(seckey_filename)) {
            fprintf(stderr, "[ERROR] Secret key %s exists.\n", seckey_filename);
            exit(EXIT_FAILURE);
        }
        
        generate(pubkey_filename, seckey_filename);
        
        printf("\nGenerating public key at: %s\n", pubkey_filename);
        printf("Generating secret key at: %s\n\n", seckey_filename);

        printf("\nUse the following command to sign a file:\n");
        printf("\n%s -S <file> -k %s [-x <sig_file>]\n\n", argv[0], seckey_filename);

        printf("\nUse the following command to verify a file signed by %s:\n", seckey_filename);
        printf("\n%s -V <file> -p %s\n\n", argv[0], pubkey_filename);
        
    } else if (sign_flag) {
        if (!sign_filename) {
            fprintf(stderr, "[ERROR] Missing filename for signing after option -s or --sign\n");
            usage();
            exit(EXIT_FAILURE);
        }
        if (!seckey_filename) {
            seckey_filename = DEFAULT_SECKEY_FILE;
            if (file_exists(seckey_filename)) {
                printf("[INFO] Using default secret key: %s\n", seckey_filename);
            } else {
                fprintf(stderr, "[ERROR] Default secret key file not found. Missing required option -k or --seckey for verifying.\n");
                usage();
                exit(EXIT_FAILURE);
            }
        }

        SecKey seckey;
        memset(&seckey, 0, sizeof(SecKey));
        load_sk_from_file(seckey_filename, &seckey);

        if (!signature_filename) {
            signature_filename = append_sig_extension(sign_filename);
            if (file_exists(signature_filename)) {
                fprintf(stderr, "[ERROR] Sig file %s already exists.\n", signature_filename);
                free(signature_filename);
                signature_filename = NULL;
                exit(EXIT_FAILURE);
            }
            free_flag = 1;
        }

        printf("[INFO] Signing file '%s' with secret key '%s'\n", sign_filename, seckey_filename);

        if (!sign_file(&seckey, sign_filename, signature_filename)) {
            fprintf(stderr, "[ERROR] Sign signature error.\n");
        } else {
            printf("[SUCCESS] Output signature file: %s\n", signature_filename);
        }

        if (free_flag) {
            free(signature_filename);
            signature_filename = NULL;
            free_flag = 0;
        }

    } else if (verify_flag) {
        if (!verify_filename) {
            fprintf(stderr, "[ERROR] Missing filename for verifying after option -v or --verify\n");
            usage();
            exit(EXIT_FAILURE);
        }
        if (!pubkey_filename) {
            pubkey_filename = DEFAULT_PUBKEY_FILE;
            if (file_exists(pubkey_filename)) {
                printf("[INFO] Using default public key: %s\n", pubkey_filename);
            } else {
                fprintf(stderr, "[ERROR] Default public key file not found. Missing required option -p or --pubkey for signing.\n");
                usage();
                exit(EXIT_FAILURE);
            }
        }

        PubKey pubkey;
        memset(&pubkey, 0, sizeof(PubKey));
        load_pk_from_file(pubkey_filename, &pubkey);

        if(!signature_filename) {
            signature_filename = append_sig_extension(verify_filename);
            if (!file_exists(signature_filename)) {
                fprintf(stderr, "[ERROR] Default sig file %s does not exist. Missing required option -x or --sigfile for verifying.\n", signature_filename);
                free(signature_filename);
                signature_filename = NULL;
                exit(EXIT_FAILURE);
            }
            free_flag = 1;
        }

        printf("[INFO] Verifying file '%s' with public key '%s' and signature '%s'\n", verify_filename, pubkey_filename, signature_filename);

        if (verify_file(&pubkey, verify_filename, signature_filename)) {
            printf("[SUCCESS] Signature verified!\n");
        } else {
            printf("[ERROR] Failed!\n");
        }

        if (free_flag) {
            free(signature_filename);
            signature_filename = NULL;
            free_flag = 0;
        }

    } else {
        usage();
        exit(EXIT_FAILURE);
    }

    return 0;
}