#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stddef.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include "utils.h"

// return 1 if file exists, otherwise return 0.
int file_exists(const char *filename) {
    struct stat buffer;
    return (stat(filename, &buffer) == 0);
}

size_t get_file_size(const char *filename) {
    struct stat st;
    if (stat(filename, &st) != 0) {
        exit_with_perror_msg("[ERROR] stat");
    }
    return st.st_size;
}

size_t get_file_size_with_fd(int fd) {
    struct stat st;
    if (fstat(fd, &st) == -1) {
        close(fd);
        exit_with_perror_msg("[ERROR] fstat");
    }
    return st.st_size;
}

void exit_with_error_msg(const char* msg) 
{
    if(msg) {
        fprintf(stderr, "[ERROR] %s.\n", msg);
    } else {
        fprintf(stderr, "[ERROR].\n");
    }
    exit(-1);
}

void exit_with_perror_msg(const char* msg) 
{
    if(msg) {
        perror(msg);
    } else {
        perror("");
    }
    exit(-1);
}

// return ERROR CODE (neg num) if failed, return 1 if succeed.
int write_file(const char *filename, char *write_content, size_t size) 
{

    int fd = open(filename, O_RDWR | O_CREAT | O_TRUNC, S_IRUSR | S_IWUSR);
    if (fd == -1) {
        fprintf(stderr, "[ERROR] Failed to open file %s's for writing.\n", filename);
        return PQSIGN_FILE_OPEN_ERROR;
    }

    if (get_file_size_with_fd(fd) < size) {
        // file size must be at least equal to size.
        if (ftruncate(fd, size) == -1) {
            perror("[ERROR] Failed to resize file");
            close(fd);
            return 0;
        }
    }

    // map the file.
    char *map = mmap(NULL, size, PROT_WRITE, MAP_SHARED, fd, 0);
    if (map == MAP_FAILED) {
        fprintf(stderr, "[ERROR] Failed to map file %s\n", filename);
        close(fd);
        return PQSIGN_FILE_MMAP_ERROR;
    }

    // write data to the file.
    memcpy(map, write_content, size);

    // unmap the file.
    if (munmap(map, size) == -1) {
        fprintf(stderr, "[ERROR] Failed to unmap memory for file %s\n", filename);
        close(fd);
        return PQSIGN_FILE_MUNMAP_ERROR;
    }

    close(fd);

    return 1;
}

// return ERROR CODE (neg num) if failed, return read_size if succeed.
int read_file(const char *filename, char *read_content, size_t max_size) 
{

    int fd = open(filename, O_RDONLY);
    if (fd == -1) {
        fprintf(stderr, "[ERROR] Failed to open file %s's for reading.\n", filename);
        return PQSIGN_FILE_OPEN_ERROR;
    }

    size_t file_size = get_file_size_with_fd(fd);

    if (file_size > max_size) {
        fprintf(stderr, "[ERROR] File %s's size is larger than expected max size.\n", filename);
        close(fd);
        return PQSIGN_FILE_SIZE_ERROR;
    }

    // map the file.
    char *map = mmap(NULL, file_size, PROT_READ, MAP_PRIVATE, fd, 0);
    if (map == MAP_FAILED) {
        fprintf(stderr, "[ERROR] Failed to map file %s\n", filename);
        close(fd);
        return PQSIGN_FILE_MMAP_ERROR;
    }

    // read data from file map.
    memcpy(read_content, map, file_size);

    // unmap the file.
    if (munmap(map, file_size) == -1) {
        fprintf(stderr, "[ERROR] Failed to unmap memory for file %s\n", filename);
        close(fd);
        return PQSIGN_FILE_MUNMAP_ERROR;
    }

    if (file_size == max_size) {
        // Ignore the last '\n'
        if (read_content[file_size - 1] == '\n') {
            read_content[file_size - 1] = '\0';
            file_size -= 1;
        }
    }

    close(fd);

    return file_size;
}
